
//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#ifdef _WIN32
#include <windows.h>
#include <windef.h>
#else
#	include <unistd.h>
#	include <sys/times.h>
#	include <sys/time.h>
#	include <string.h>
#endif

#ifdef __APPLE__
#	include <CoreServices/CoreServices.h>
#	include <mach/mach.h>
#	include <mach/mach_time.h>
#endif


#include "NanoTimer.h"



/**  Get the user time in milliseconds
 *
 * @return
 *   Time in milliseconds.
 *
 * The time is the user time of the process (and all children) since the
 * process started.
 *
 * @implementation
 *   Uses @c times under Unix, and @c GetProcessTimes under Windoze.
 *   Warning: Under Windoze, clock() returns wall-clock time, *not*
 *   user time! (contrary to what the MSDN documentation says!)
 *
 **/

float time_ms( void )
{
#ifdef _WIN32

	unsigned long long utime;
	FILETIME creation_time, exit_time, kernel_time, user_time;
	bool success = GetProcessTimes( GetCurrentProcess(), &creation_time, &exit_time,
									&kernel_time, &user_time);
	if ( ! success )
		return 0.0;
	utime =   ( static_cast<unsigned long long>(user_time.dwHighDateTime) << 32) 
			+ static_cast<unsigned>(user_time.dwLowDateTime);
	return static_cast<float>( utime ) / 1.0e4f;

#else

	struct tms t;
	times( &t );
	return static_cast<float>(t.tms_utime) * 1000.0f / sysconf(_SC_CLK_TCK);

#endif
}





/** @class NanoTimer
 *
 * Timer with nanoseconds resolution.
 *
 * The units of this timer are always nanoseconds,
 * but on some platforms, the actual resolution might be less (microseconds or
 * even less).
 *
 * Where implemented, this class uses the high-speed, high-performance, hardware
 * timers/counters.
 * Otherwise, we just use @a gettimeofday(), which, at least on Linux/Pentium,
 * has microsecond resolution.
 *
 * Currently, the time is @e wall-clock time, not user time!
 *
 * @see
 * - http://www.ncsa.uiuc.edu/UserInfo/Resources/Hardware/IA32LinuxCluster/Doc/timing.html
 * - cedar.intel.com/software/idap/media/pdf/rdtscpm1.pdf
 * - /raphael/knowledge/programming/c/code-snippets/cpu_clock_timer.c
 *
 * @todo
 * - Try to estimate the overhead of a function call to start()
 *   or elapsed().
 * - Use PAPI (http://icl.cs.utk.edu/projects/papi/)
 *   or the "high resolution timers project"
 *   (http://high-res-timers.sourceforge.net/)
 *   when they become widely available (without kernel patches).
 *
 **/


/**  Create a new timer with nanoseconds resolution
 *
 * Saves the current time stamp, too.
 *
 * @sideeffects
 *   See @a checkFrequency().
 *
 **/

NanoTimer::NanoTimer(void)
	: m_first_time(true)
{
	for ( unsigned int i=0; i < NANO_TIMER_NUM_HISTORY-1; i ++ )
		m_history[i] = 0;

	if ( ! M_FrequencyChecked )
		checkFrequency();
	start();
}



/**  Save the current time (stamp) in the timer
 *
 **/

void NanoTimer::start( void )
{
	m_time_stamp = getTimeStamp();
}



/**  Return the time elapsed since the last @a start() in nanoseconds.
 *
 **/

double NanoTimer::elapsed( void ) const
{
	unsigned long long int now = getTimeStamp();
	return convert( now - m_time_stamp );
}




/**  Save time elapsed since last start in history.
*
**/

void NanoTimer::stop( void )
{
	double intvl = elapsed();
	
	if ( m_first_time )
	{
		m_first_time = false;
		for ( unsigned int i=0; i < NANO_TIMER_NUM_HISTORY; i ++ )
			m_history[i] = intvl;
		return;
	}
	
	for ( unsigned int i=0; i < NANO_TIMER_NUM_HISTORY-1; i ++ )
		m_history[i] = m_history[i+1];
	m_history[NANO_TIMER_NUM_HISTORY-1] = intvl;
}



/**  Return the average of the time history in nanoseconds.
*
**/

double NanoTimer::average( void )
{
	double s = 0.0;
	for ( unsigned int i=0; i < NANO_TIMER_NUM_HISTORY-1; i ++ )
		s += m_history[i];
	return s /= NANO_TIMER_NUM_HISTORY;
}



/** Convert internal interval time format to nanoseconds.
 *
 **/

double NanoTimer::convert( unsigned long long int interval )
{
	if ( M_Use_High_Frequ )
		return static_cast<double>(interval) / M_GHz;
	else
		return static_cast<double>(interval) * 1000.0;
}



/**  Return current time stamp
 *
 * Should be used only internally by this class,
 * because the meaning of the return value is different depending on whether
 * @a M_Use_High_Frequ is true or not.
 *
 * @warning
 *   We don't check whether checkFrequency succeded!
 *   If it failed, the times will be bogus, or, in the worst case, the program
 *   might even crash.
 *
 * @implementation
 *   Depending on the platform where this is compiled, different code is
 *   compiled. Be careful about the correct order of the #ifdef's! 
 *   For the apple stuff, see http://developer.apple.com/qa/qa2004/qa1398.html .
 **/

#if defined(_WIN32)
#pragma message( "WARNING: NanoTimer will use QueryPerformanceCounter." )  

long long unsigned int NanoTimer::getTimeStamp( void )
{
	LARGE_INTEGER stamp;
	QueryPerformanceCounter( &stamp );
	return stamp.QuadPart;
}

#elif defined(__APPLE__)
#warning "NanoTimer will use Microseconds()."

long long unsigned int NanoTimer::getTimeStamp( void )
{
	// from http://developer.apple.com/qa/qa2004/qa1398.html
	uint64_t        stamp;
    Nanoseconds     nano;
		
    stamp = mach_absolute_time();
	
    // Convert to nanoseconds.	
    // (Have to do some pointer fun because AbsoluteToNanoseconds
    // works in terms of UnsignedWide, which is a structure rather
    // than a proper 64-bit integer.)
    nano = AbsoluteToNanoseconds( reinterpret_cast<AbsoluteTime &>(stamp) );
	
    return reinterpret_cast<uint64_t &>( nano );
}

#elif defined(__GNUC__) && defined(__linux)
#warning "NanoTimer will use pentium assembler code"

long long unsigned int NanoTimer::getTimeStamp( void )
{
	unsigned long long int timestamp;
	__asm__ volatile ("rdtsc" : "=A" (timestamp) : );
	return timestamp;

}

#else // all platforms not yet implemented
#warning "NanoTimer will use col::time_ms() !!"

long long unsigned int NanoTimer::getTimeStamp( void )
{
	return col::time_ms() * 1000;
}

#endif



/**  Determine the clock frequency of the CPU
 *
 * Try to determine the speed at which the CPUs time-stamp counter is running
 * (on those plastforms where NanoTimer uses this counter),
 * and whether or not there is a high-frequency counter at all.
 *
 * @pre
 *   Assumes that all CPUs in one system run with the same clock frequency!
 *
 * @sideeffects
 *   Class variable M_GHz.
 *
 * @implementation
 *   Under Linux, we parse /proc/cpuinfo; under Windoze, we use QueryPerformanceFrequency.
 *
 * @see
 *   getTimeStamp(). The #ifdef's must match!
 * 
 **/

void NanoTimer::checkFrequency( void )
{
	M_GHz = 0.001;
	M_Use_High_Frequ = false;

#if defined( __APPLE__ )

	M_Use_High_Frequ = true;
	M_GHz = 1.0;
	
#elif defined(__linux)
	FILE *cpuinfo = fopen("/proc/cpuinfo", "r");
	if ( ! cpuinfo )
	{
		perror("fopen(/proc/cpuinfo)");
		fputs("col::NanoTimer: Can't determine CPU frequency!\n",stderr);
	}
	else
	{
		char cpuinfo_buf[1000];

		while ( fgets(cpuinfo_buf, sizeof(cpuinfo_buf), cpuinfo) )
		{
			if ( ! strncasecmp(cpuinfo_buf, "cpu mhz", 7) )
			{
				double mhz;
				int assigned;

				assigned = sscanf(cpuinfo_buf, "%*s %*s : %lf", &mhz);

				if ( assigned != 1 )
					fputs("col::NanoTimer: Can't parse /proc/cpuinfo !\n",
						  stderr );
				else
				{
					M_GHz = mhz / 1000.0;
					M_Use_High_Frequ = true;
				}

				break;
			}
		}

		fclose( cpuinfo );
	}

#elif defined( _WIN32 )

	LARGE_INTEGER hz;
	bool use_high_frequ = QueryPerformanceFrequency( &hz );
	if ( ! use_high_frequ )
		fputs("col::NanoTimer: Can't use high-frequency counter!\n",stderr);
	else
	{
		M_GHz = hz.QuadPart / 1.0E9;
		M_Use_High_Frequ = true;
	}

#endif

	if ( M_Use_High_Frequ )
		fprintf(stderr, "col::NanoTimer: found %lf GHz\n", M_GHz );
	else
	{
		puts("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
		puts("col::NanoTimer: will be bogus !!");
		puts("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
	}

	M_FrequencyChecked = true;
}



/**  Tells whether or not the NanoTimer use the high frequency counter
 *
 * @warning
 *   Valid only after the first NanoTimer has been created!
 **/

bool NanoTimer::usesHighFrequ( void )
{
	return M_Use_High_Frequ;
}



/**  Returns the frequency (resolution) of the counter in GHz
 *
 * @warning
 *   Valid only if @a NanoTimer::usesHighFrequ() = @a true !
 **/

double NanoTimer::frequ( void )
{
	return M_GHz;
}


bool	NanoTimer::M_Use_High_Frequ		= false;
double	NanoTimer::M_GHz				= 0.001;
bool	NanoTimer::M_FrequencyChecked	= false;


